/*
 * main.cpp
 *
 * Copyright (C) 2003 Ed Cogburn <ecogburn@xtn.net>
 *
 * This file is part of the program "d: The Directory Lister".
 *
 * "d: The Directory Lister" is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 *
 * "d: The Directory Lister" is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with "d: The Directory Lister" in a file named "COPYING".
 * If not, visit their website at http://www.gnu.org for a copy, or request a copy by writing to:
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
*/



#include "globals.hpp"
#include <sys/types.h>
#include <dirent.h>
#include <ctime>
#include <cmath>
#include <fnmatch.h>



// Option numbers used to identify options passed to parse_options()
enum opt_nbr {
	onEndOfList = 0, onColor, onColorize, onDirsFirst, onHelp,
	onKilobyteReal, onPadding, onReverseSort, onSort, onVersion
};

// Option definitions passed to parse_options()
// Note the last entry must have a 0 as the first number (onEndOfList in this case)
po_declared_option opts[] = {
	{ onColor,			poasCfgOnly,	poatString,		'\0',	"color"			},
	{ onColorize,		poasOptCfg,		poatBoolean,	'c',	"colorize"		},
	{ onDirsFirst,		poasOptCfg,		poatBoolean,	'd',	"dirs-first"	},
	{ onHelp,			poasOptOnly,	poatNone,		'h',	"help"			},
	{ onKilobyteReal,	poasOptCfg,		poatBoolean,	'k',	"kilobyte-real"	},
	{ onPadding,		poasOptCfg,		poatInt,		'p',	"padding"		},
	{ onReverseSort,	poasOptCfg,		poatBoolean,	'r',	"reverse-sort"	},
	{ onSort,			poasOptOnly,	poatString,		's',	"sort"			},
	{ onVersion,		poasOptOnly,	poatNone,		'v',	"version"		},
	{ onEndOfList,		poasVoid,		poatVoid,		'\0',	""				}
};

parse_options po;

colorized_text ct;

// Our option flags
bool oHelp = false;
bool oVersion = false;
bool oColorize = true;
bool oDirsFirst = true;
bool oKilobyteReal = true;
bool oReverseSort = false;
string oSort = "name";
int oPadding = 2;

// These control how much space must be reserved for their section.  They're calculated on a first pass before printing.
unsigned uid_name_section_width;
unsigned gid_name_section_width;
unsigned file_size_section_width;

char* vers = "1.0.0";


// This stuff is for colorization output.
struct color_setting
{
	ctColor fg;
	ctColor bg;
	ctAttr at;
};

// Default colorizations used by d.  This array is indexed by colorize_index below.
color_setting colorize[] = {
	{  ctcYellow,   ctcRed,   ctaBold },
	{    ctcCyan, ctcBlack,   ctaBold },
	{   ctcGreen, ctcBlack,   ctaBold },
	{ ctcMagenta, ctcBlack,   ctaBold },
	{   ctcWhite, ctcBlack, ctaNormal },
	{  ctcYellow,  ctcBlue,   ctaBold },
	{   ctcWhite,  ctcBlue,   ctaBold },
	{     ctcRed,  ctcBlue,   ctaBold }
};

enum colorize_index { ciBadLink, ciDir, ciExec, ciLink, ciNormal, ciChrDev, ciBlkDev, ciSpecial };

// This struct is for user provided colorizations.
struct custom_color_setting
{
	string name;
	ctColor fg;
	ctColor bg;
	ctAttr at;
};

// This will hold the user provided colorizations, after being parsed.
vector<custom_color_setting> filecolorlist;



//-----------------------------------------------------------------------------
void print_vers_msg(stringstream& ss)
{
	ss << "d v" << vers << endl;
}



//-----------------------------------------------------------------------------
// yes, 2 different functions for printing the help message is ugly, but colorizing the output 
// complicates the source as you can see.
void print_help_msg_color(stringstream& ss)
{
	ss << ct.set(ctcYellow, ctcBlack, ctaBold);
	ss << "d:  The Directory Lister.  Version " << vers << endl;

	ss << ct.set(ctcWhite, ctcBlack, ctaBold);
	ss << "Copyright @2003 Ed Cogburn.  Released under the GNU General Public License." << endl;
	ss << "See the file COPYING for details.  There is NO warranty, use at your own risk." << endl << endl;

	ss << ct.reset();

	ss << "Usage:  d [options] [path]" << endl << endl;

	ss << "Options:\n" << endl;

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--help" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-h" << ct.reset();
	ss << "\t\t\tThis help screen.\n";

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--version" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-v" << ct.reset();
	ss << "\t\tVersion number of 'd'." << endl;

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--colorize" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-c" << ct.reset();
	ss << "\t\tColorize output.  Default: ON." << endl;

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--dirs-first" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-d" << ct.reset();
	ss << "\t\tAlways print directories first.  Default: ON." << endl;

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--kilobyte-real" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-k" << ct.reset();
	ss << "\t\tUse a real kilobyte, 1,024, instead of 1,000" << endl << "\t\t\t\tfor calculating file size.  Default: ON.\n";

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--reverse-sort" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-r" << ct.reset();
	ss << "\t\tPrint output in reverse order.  Default: OFF." << endl;

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--padding=" << ct.reset();
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "arg" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-p=" << ct.reset();
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "arg" << ct.reset();
	ss << "\tSet spacing between sections on a line"  << endl << "\t\t\t\t(1<=arg<=16).  Default: 2.\n";

	ss << "  ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "--sort=";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "arg" << ct.reset();
	ss << " or ";
	ss << ct.set(ctcGreen, ctcBlack, ctaBold) << "-s=";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "arg" << ct.reset();
	ss << "\t\tSet sorting method.  Arg can be '(";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "d" << ct.reset();
	ss << ")";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "ate" << ct.reset();
	ss << "',"  << endl << "\t\t\t\t'(";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "n" << ct.reset();
	ss << ")";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "ame" << ct.reset();
	ss << "' or '(";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "s" << ct.reset();
	ss << ")";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "ize" << ct.reset();
	ss << "'.  Default: '";
	ss << ct.set(ctcMagenta, ctcBlack, ctaBold) << "name" << ct.reset();
	ss << "'." << endl << endl;
		
	ss << "Use '--option=false' or '-o-' to turn an option off, or use the /etc/default/d or" << endl;
	ss << "/etc/d.conf or $HOME/.d.conf configuration file.  See docs for details." << endl;
}



//-----------------------------------------------------------------------------
// Nice and short, but looks dull.
void print_help_msg_nocolor(stringstream& ss)
{
	ss << "The Directory Lister.  Version " << vers << endl;

	ss << "Copyright @2003 Ed Cogburn.  Released under the GNU General Public License." << endl;
	ss << "See the file COPYING for details.  There is NO warranty, use at your own risk." << endl << endl;

	ss << "Usage:  d [options] [path]" << endl << endl;

	ss << "Options:" << endl;

	ss << "  --help or -h\t\t\tThis help screen." << endl;
	ss << "  --version or -v\t\tVersion number of 'd'." << endl;
	ss << "  --colorize or -c\t\tColorize output.  Default: ON." << endl;
	ss << "  --dirs-first or -d\t\tAlways print directories first.  Default: ON." << endl;
	ss << "  --kilobyte-real or -k\t\tUse a real kilobyte, 1,024, instead of 1,000" << endl;
	ss <<        "\t\t\t\tfor calculating file size.  Default: ON." << endl;
	ss << "  --reverse-sort or -r\t\tPrint output in reverse order.  Default: OFF." << endl;
	ss << "  --padding=arg or -p=arg\tSet spacing between sections on a line" << endl;
	ss <<        "\t\t\t\t(1<=arg<=16).  Default: 2." << endl;
	ss << "  --sort=arg or -s=arg\t\tSet sorting method.  Arg can be '(d)ate'," << endl;
	ss <<        "\t\t\t\t'(n)ame' or '(s)ize'.  Default: 'name'." << endl << endl;

	ss << "Use '--option=false' or '-o-' to turn an option off, or use the /etc/default/d or" << endl;
	ss << "/etc/d.conf or $HOME/.d.conf configuration file.  See docs for details." << endl;
}



//-----------------------------------------------------------------------------
// Used in more than one place, by pring_file_size, and to calculate total_size of all files.
static string human_readable_file_size(unsigned long size)
{
	stringstream ss;
	long double val = size;
	long double divisor;
	char k, m, g, t, e;

	if (oKilobyteReal) {
		divisor = 1024;
		k = 'K';
		m = 'M';
		g = 'G';
		t = 'T';
		e = 'E';
	} else {
		divisor = 1000;
		k = 'k';
		m = 'm';
		g = 'g';
		t = 't';
		e = 'e';
	}

	// For comments below I use 'k', 'm', 'g', 't' and 'e' but they could also be uppercase if oKilobyteReal == true.

	// Width = NNNN | ( (N.N | NN | 0NNN) + (k | m) )
	ss << fixed;

	// Less than 1000/1024?
	if (val < divisor) {
		ss << setprecision(0) << val;
		return ss.str();
	}

	// First division, now = N + k
	val /= divisor;

	ss << setprecision(0) << setw(file_size_section_width - 1);

	// Less than 10k?
	if (val < 10) {
		// N.N + k
		// Round up or down if within 0.1
		if (val - floor(val) > 0.9)
			val = floor(val) + 1.0;
		if (val - floor(val) >= 0.1)
			ss << setprecision(1);
		ss << val << k;
		return ss.str();
	}

	// Less than 100k?
	if (val < 100) {
		// NN + k
		ss << val << k;
		return ss.str();
	}

	// Less than 1000k?
	if (val < 1000) {
		// NNN + k
		ss << val << k;
		return ss.str();
	}

	val /= divisor;

	if (val < 10) {
		// N.N + m
		// Round up or down if within 0.1
		if (val - floor(val) > 0.9)
			val = floor(val) + 1.0;
		if (val - floor(val) >= 0.1)
			ss << setprecision(1);
		ss << val << m;
		return ss.str();
	}

	if (val < 100) {
		// NN + m
		ss << val << m;
		return ss.str();
	}

	if (val < 1000) {
		// NNN + m
		ss << val << m;
		return ss.str();
	}

	val /= divisor;

	if (val < 10) {
		// N.N + g
		// Round up or down if within 0.1
		if (val - floor(val) > 0.9)
			val = floor(val) + 1.0;
		if (val - floor(val) >= 0.1)
			ss << setprecision(1);
		ss << val << g;
		return ss.str();
	}

	if (val < 100) {
		// NN + g
		ss << val << g;
		return ss.str();
	}

	if (val < 1000) {
		// NNN + g
		ss << val << g;
		return ss.str();
	}

	val /= divisor;

	if (val < 10) {
		// N.N + t
		// Round up or down if within 0.1
		if (val - floor(val) > 0.9)
			val = floor(val) + 1.0;
		if (val - floor(val) >= 0.1)
			ss << setprecision(1);
		ss << val << t;
		return ss.str();
	}

	if (val < 100) {
		// NN + t
		ss << val << t;
		return ss.str();
	}

	// if (val < 1000) {
	// NNN + t
	ss << val << t;
	return ss.str();

	// Does anyone have files bigger than 999 terabytes?  :-)
}



//-----------------------------------------------------------------------------
// An encapsulation of the process of printing a file's information on one line.
class print_file_line
{
public:
	print_file_line(file_info& f, int padding);
	string str() { return pfl_ss.str(); };

protected:
	string permissions(file_info& f);
	string name(file_info& f, bool include_dir);
	string date(file_info& f);

	stringstream pfl_ss;
};



//-----------------------------------------------------------------------------
// ctor does all the work.
print_file_line::print_file_line(file_info& f, int padding)
{
	string pad;

	// Set padding for output lines
	for (int i = 0; i < padding; i++)
		pad += " ";

	pfl_ss << permissions(f);

	pfl_ss << pad;

	// Owner/Group names
	pfl_ss << setw(uid_name_section_width) << f.uid_name << ':';
	pfl_ss << setw(gid_name_section_width) << left << f.gid_name << right << setw(0);

	pfl_ss << pad;

	// Either file size or major/minor if a device
	pfl_ss << setw(file_size_section_width);
	if (f.is_char_device || f.is_block_device) {
		stringstream tmp_ss;
		tmp_ss << ((f.rdev_nbr >> 8) & 0xFF) << '-' << (f.rdev_nbr & 0xFF);
		pfl_ss << tmp_ss.str();
	} else
		pfl_ss << human_readable_file_size(f.size);
	pfl_ss << setw(0);

	pfl_ss << pad;

	pfl_ss << date(f);

	pfl_ss << pad;

	pfl_ss << name(f, false);

	// Print name of file being pointed to if a symbolic link
	if (f.is_symlink) {
		path fp(f.name.dir(), f.symlink_name);
		file_info ff(fp.full().c_str());

		pfl_ss << " -> ";
		if (f.name.dir() != ff.name.dir())
			pfl_ss << name(ff, true);
		else
			pfl_ss << name(ff, false);
	}

	pfl_ss << '\n';
}



//-----------------------------------------------------------------------------
// build the "rwxrwxrwx" string
string print_file_line::permissions(file_info& f)
{
	char t;
	string perms;

	char d = '-';
	char r = 'r';
	char w = 'w';
	char x = 'x';

	if (f.perm.owner_r)
		t = r;
	else
		t = d;
	perms += t;

	if (f.perm.owner_w)
		t = w;
	else
		t = d;
	perms += t;

	if (f.perm.suid_x)
		t = 's';
	else
		if (f.perm.owner_x)
			t = x;
		else
			t = d;
	perms += t;

	if (f.perm.group_r)
		t = r;
	else
		t = d;
	perms += t;

	if (f.perm.group_w)
		t = w;
	else
		t = d;
	perms += t;

	if (f.perm.sgid_x)
		t = 's';
	else
		if (f.perm.group_x)
			t = x;
		else
			t = d;
	perms += t;

	if (f.perm.other_r)
		t = r;
	else
		t = d;
	perms += t;

	if (f.perm.other_w)
		t = w;
	else
		t = d;
	perms += t;

	if (f.perm.sticky)
		t = 't';
	else
		if (f.perm.other_x)
			t = x;
		else
			t = d;
	perms += t;

	return perms;
}



//-----------------------------------------------------------------------------
// Print date much like ls does.
string print_file_line::date(file_info& f)
{
	string pfl_date;
	time_t T;
	struct tm dt;
	double tdiff;
	char buf[64];

	// first break down file time
	if (localtime_r(&f.mod_time, &dt) != &dt) {
		stringstream err;
		err << "print_file_date(): Can't convert time-stamp for file \"" << f.name.name_ext() << "\"!";
		throw runtime_error(err.str());
	}

	T = time(NULL);
	tdiff = difftime(T, f.mod_time);
	if (tdiff < 14515200.0) {   // minimum number of seconds in 6 months (28 days per month)
		// print month, day, time
		strftime(buf, 64, "%b %d %H:%M", &dt);
	} else {
		// print month, day, year
		strftime(buf, 64, "%b %d  %Y", &dt);
	}
	pfl_date += buf;

	return pfl_date;
}



//-----------------------------------------------------------------------------
// Printing the name is made complicated by colorization.
string print_file_line::name(file_info& f, bool include_dir)
{
	string pfl_name;

	if (oColorize && f.exists)
		// If all execute bits set then check for directory or symlink first.
		if (f.perm.owner_x || f.perm.group_x || f.perm.other_x)
			if (f.is_dir)
				pfl_name = ct.set(colorize[ciDir].fg, colorize[ciDir].bg, colorize[ciDir].at);
			else
				if (f.is_symlink)
					if (f.is_dangling_symlink)
						pfl_name = ct.set(colorize[ciBadLink].fg, colorize[ciBadLink].bg, colorize[ciBadLink].at);
					else
						pfl_name = ct.set(colorize[ciLink].fg, colorize[ciLink].bg, colorize[ciLink].at);
				else
					pfl_name = ct.set(colorize[ciExec].fg, colorize[ciExec].bg, colorize[ciExec].at);
		else
			// Look for specials
			if (f.is_pipe || f.is_socket)
				pfl_name = ct.set(colorize[ciSpecial].fg, colorize[ciSpecial].bg, colorize[ciSpecial].at);
			else
				if (f.is_char_device)
					pfl_name = ct.set(colorize[ciChrDev].fg, colorize[ciChrDev].bg, colorize[ciChrDev].at);
				else
					if (f.is_block_device)
						pfl_name = ct.set(colorize[ciBlkDev].fg, colorize[ciBlkDev].bg, colorize[ciBlkDev].at);
					else
						// Check against user's list of colorizations
						for (unsigned i = 0; i < filecolorlist.size(); i++)
							if (fnmatch(filecolorlist[i].name.c_str(), f.name.name_ext().c_str(), 0) == 0) {
								pfl_name = ct.set(filecolorlist[i].fg, filecolorlist[i].bg, filecolorlist[i].at);
								break;
							}

	// If not selected above, must be normal file.
	if (oColorize && pfl_name.size() == 0)
		pfl_name = ct.set(colorize[ciNormal].fg, colorize[ciNormal].bg, colorize[ciNormal].at);

	if (include_dir)
		pfl_name += f.name.full();
	else
		pfl_name += f.name.name_ext();

	if (f.is_dir)
		pfl_name += "/";

	if (oColorize)
		pfl_name += ct.reset();

	return pfl_name;
}



//-----------------------------------------------------------------------------
// This is one of the "Comp" functions used for sorting the list.
// If you're wondering, no, I wouldn't have known how to do this without a reference book.  :-)
// If you want to know what this is, try "C++ The Complete Reference", 4th Ed,
// by Herbert Schildt, Chapter 24, "Using Function Objects".
class sort_list_compare_files_by_date : binary_function<file_info&, file_info&, bool>
{
public:
	result_type operator() (first_argument_type arg1, second_argument_type arg2)
	{ 
		if (arg1.mod_time == arg2.mod_time)
			return static_cast<result_type> (arg1.name.full() < arg2.name.full());
		else
			return static_cast<result_type> (arg1.mod_time < arg2.mod_time);
	}
};


		
//-----------------------------------------------------------------------------
// This is one of the "Comp" functions used for sorting the list.
class sort_list_compare_files_by_name : binary_function<file_info&, file_info&, bool>
{
public:
	result_type operator() (first_argument_type arg1, second_argument_type arg2)
	{ 
		return static_cast<result_type> (arg1.name.full() < arg2.name.full());
	}
};


		
//-----------------------------------------------------------------------------
// This is one of the "Comp" functions used for sorting the list.
class sort_list_compare_files_by_size : binary_function<file_info&, file_info&, bool>
{
public:
	result_type operator() (first_argument_type arg1, second_argument_type arg2)
	{ 
		if (arg1.size == arg2.size)
			return static_cast<result_type> (arg1.name.full() < arg2.name.full());
		else
			return static_cast<result_type> (arg1.size < arg2.size);
	}
};


		
//-----------------------------------------------------------------------------
void dump_list(list<file_info>& flist)
{
	unsigned long total_size;
	int nbr_files;
	vector<string> output_lines;

    // All the sorting work is done by list<> itself and the "Comp" functions above.
	if (flist.size() > 1)
		if (oSort == "date")
			flist.sort(sort_list_compare_files_by_date());
		else
			if (oSort == "name")
				flist.sort(sort_list_compare_files_by_name());
			else
				if (oSort == "size")
					flist.sort(sort_list_compare_files_by_size());
				else
					throw logic_error("sort_list(): Illegal oSort value!");

	// Bubble sort directories to the top of the list	
	if (oDirsFirst && flist.size() > 1) {
		bool Cont;
		do {
			Cont = false;
			list<file_info>::iterator it1 = flist.begin();
			it1++;
			while (it1 != flist.end()) {
				list<file_info>::iterator it2 = it1;
				it2--;
				if ( (*it1).is_dir && !(*it2).is_dir ) {
					swap(*it1, *it2);
					Cont = true;
				}
				it1++;
			}
		} while (Cont == true);
	}

	// initialize vars
	file_size_section_width = 4;
	uid_name_section_width = 0;
	gid_name_section_width = 0;
	total_size = 0;
	nbr_files = 0;

	// This is our first pass to calculate certain vars for optimizing output on the second pass.
	// Check size of user and group names.
	// Check whether devices are being displayed to decide amount of space needed for the file-size/device-major-minor section.
	// Add up number of normal files and their total size.
	for (list<file_info>::iterator it = flist.begin(); it != flist.end(); it++) {
		unsigned tmp;

		// Track the maximum these vars need to be
		tmp = (*it).uid_name.size();
		if (tmp > uid_name_section_width)
			uid_name_section_width = tmp;
		tmp = (*it).gid_name.size();
		if (tmp > gid_name_section_width)
			gid_name_section_width = tmp;

		// If displaying char/block devices we need more space for that output.
		if ((*it).is_char_device || (*it).is_block_device)
			file_size_section_width = 7;

		// For the one line summary at the end of the printout.
		if ((*it).is_regular_file) {
			nbr_files++;
			total_size += (*it).size;
		}
	}

	// Reverse sort?
	if (oReverseSort)
		reverse(flist.begin(), flist.end());

	// Go thru the list and create the output lines
	for (list<file_info>::iterator it = flist.begin(); it != flist.end(); it++) {
		print_file_line pfl((*it), oPadding);
		output_lines.push_back(pfl.str());
	}

	// Dump the lines
	for (unsigned i = 0; i < output_lines.size(); i++)
		cout << output_lines[i];

	// One line summary
	if (nbr_files)
		cout << nbr_files << " regular files, " <<  human_readable_file_size(total_size) << " total size." << endl;
}



//-----------------------------------------------------------------------------
// Use libc's opendir/readdir to suck in a directory's contents.
void dump_directory(list<file_info>& flist, file_info& df)
{
	struct dirent *de;
	DIR *d = opendir(df.name.full().c_str());
	string s;

	flist.clear();

	while (d != NULL) {
		de = readdir(d);
		if (de == NULL)
			break;

		if (de->d_name[0] == '.') {
			if (de->d_name[1] == '\0')
				continue;
			if (de->d_name[1] == '.' && de->d_name[2] == '\0')
				continue;
		}

		s = df.name.full();
		s += '/';
		s += de->d_name;

		file_info f = s;
		flist.push_back(f);
	}
}



//-----------------------------------------------------------------------------
// Part of parsing a "color" option in a configuration file.
ctColor read_config_validate_color(string& c, const char* filename, unsigned line)
{
	while (c[0] == ' ' || c[0] == '\t')
		c.erase(0, 1);

	while (c[c.size() - 1] == ' ' || c[c.size() - 1] == '\t')
		c.erase(c.size() - 1, 1);

	if (c == "black")
		return ctcBlack;
	if (c == "red")
		return ctcRed;
	if (c == "green")
		return ctcGreen;
	if (c == "yellow")
		return ctcYellow;
	if (c == "blue")
		return ctcBlue;
	if (c == "magenta")
		return ctcMagenta;
	if (c == "cyan")
		return ctcCyan;
	if (c == "white")
		return ctcWhite;

	stringstream err;
	err << "validate_color(): Invalid color \"" << c << "\" given in \"" << filename << "\" at line " << line << '!';
	err << "  Must be 'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', or 'white'.";
	throw runtime_error(err.str());
}



//-----------------------------------------------------------------------------
// Part of parsing a "color" option in a configuration file.
ctAttr read_config_validate_attr(string& a, const char* filename, unsigned line)
{
	while (a[0] == ' ' || a[0] == '\t')
		a.erase(0, 1);

	while (a[a.size() - 1] == ' ' || a[a.size() - 1] == '\t')
		a.erase(a.size() - 1, 1);

	if (a == "normal")
		return ctaNormal;
	if (a == "bold")
		return ctaBold;
	if (a == "underline")
		return ctaUnderline;
	if (a == "blink")
		return ctaBlink;
	if (a == "reverse")
		return ctaReverse;

	stringstream err;
	err << "validate_attr(): Invalid attribute \"" << a << "\" given in \"" << filename << "\" at line " << line << '!';
	err << "  Must be 'normal', 'bold', 'underline', 'blink', or 'reverse'.";
	throw runtime_error(err.str());
}



//-----------------------------------------------------------------------------
// Part of parsing a "color" option in a configuration file.
void read_config_color(string& target, string& fcolor, string& bcolor, string& attr, const char* filename, unsigned line)
{
	while (target[0] == ' ' || target[0] == '\t')
		target.erase(0, 1);

	while (target[target.size() - 1] == ' ' || target[target.size() - 1] == '\t')
		target.erase(target.size() - 1, 1);

	if (target == "::badlink::") {
		colorize[ciBadLink].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciBadLink].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciBadLink].at = read_config_validate_attr(attr, filename, line);
		return;
	}
	if (target == "::dir::") {
		colorize[ciDir].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciDir].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciDir].at = read_config_validate_attr(attr, filename, line);
		return;
	}
	if (target == "::exec::") {
		colorize[ciExec].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciExec].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciExec].at = read_config_validate_attr(attr, filename, line);
		return;
	}
	if (target == "::link::") {
		colorize[ciLink].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciLink].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciLink].at = read_config_validate_attr(attr, filename, line);
		return;
	}
	if (target == "::normal::") {
		colorize[ciNormal].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciNormal].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciNormal].at = read_config_validate_attr(attr, filename, line);
		return;
	}
	if (target == "::chrdev::") {
		colorize[ciChrDev].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciChrDev].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciChrDev].at = read_config_validate_attr(attr, filename, line);
		return;
	}

	if (target == "::blkdev::") {
		colorize[ciBlkDev].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciBlkDev].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciBlkDev].at = read_config_validate_attr(attr, filename, line);
		return;
	}

	if (target == "::special::") {
		colorize[ciSpecial].fg = read_config_validate_color(fcolor, filename, line);
		colorize[ciSpecial].bg = read_config_validate_color(bcolor, filename, line);
		colorize[ciSpecial].at = read_config_validate_attr(attr, filename, line);
		return;
	}

	custom_color_setting ccs;

	ccs.name = target;
	ccs.fg = read_config_validate_color(fcolor, filename, line);
	ccs.bg = read_config_validate_color(bcolor, filename, line);
	ccs.at = read_config_validate_attr(attr, filename, line);

	filecolorlist.push_back(ccs);
}



//-----------------------------------------------------------------------------
// Parse a configuration file.  Parsing is simple for everything but the "color" option.
void read_config(const char* filename)
{
	po.parse_cfgs(filename, opts);

	for (unsigned i = 0; i < po.options_array.size(); i++) {
		if (po.options_array[i].nbr == onColorize) {
			oColorize = po.options_array[i].arg.bool_val;
			continue;
		}
		if (po.options_array[i].nbr == onDirsFirst) {
			oDirsFirst = po.options_array[i].arg.bool_val;
			continue;
		}
		if (po.options_array[i].nbr == onKilobyteReal) {
			oKilobyteReal = po.options_array[i].arg.bool_val;
			continue;
		}
		if (po.options_array[i].nbr == onPadding) {
			if (po.options_array[i].arg.int_val < 1 || po.options_array[i].arg.int_val > 16) {
				stringstream err;
				err << "read_config(): Invalid amount for the \"padding\" option in \"";
				err << filename << "\" at line " << po.options_array[i].lineno << '!';
				err << "  Must be between 1 and 16.";
				throw runtime_error(err.str());
			}
			oPadding = po.options_array[i].arg.int_val;
			continue;
		}
		if (po.options_array[i].nbr == onColor && oColorize == true) {
			string src = po.options_array[i].arg.string_val;
			unsigned idx;

			stringstream err;
			err << "read_config(): Invalid syntax for color option in \"";
			err << filename << "\" at line " << po.options_array[i].lineno << '!';

			idx = src.find(",", 0);
			if (idx == src.npos)
				throw runtime_error(err.str());
			string target = src.substr(0, idx);
			src.erase(0, idx + 1);

			idx = src.find(",", 0);
			if (idx == src.npos)
				throw runtime_error(err.str());
			string fcolor = src.substr(0, idx);
			src.erase(0, idx + 1);

			idx = src.find(",", 0);
			if (idx == src.npos)
				throw runtime_error(err.str());
			string bcolor = src.substr(0, idx);
			src.erase(0, idx + 1);

			if (src.size() == 0)
				throw runtime_error(err.str());
			string attr = src;

			read_config_color(target, fcolor, bcolor, attr, filename, po.options_array[i].lineno);
			continue;
		}
	}
}



//-----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
	try {

	list<file_info> filelist;

	// Get configs (system)
	file_info cf = "/etc/default/d";
	if (cf.exists)
		read_config("/etc/default/d");
	else {
		cf = "/etc/d.conf";
		if (cf.exists)
			read_config("/etc/d.conf");
	}

	// Get configs (user)
	cf = "~/.d.conf";
	if (cf.exists)
		read_config("~/.d.conf");

	// Get command line options
	po.parse_args(argc, argv, opts);
	for (unsigned i = 0; i < po.options_array.size(); i++) {
		if (po.options_array[i].nbr == onColorize) {
			oColorize = po.options_array[i].arg.bool_val;
			continue;
		}
		if (po.options_array[i].nbr == onDirsFirst) {
			oDirsFirst = po.options_array[i].arg.bool_val;
			continue;
		}
		if (po.options_array[i].nbr == onHelp) {
			oHelp = true;
			continue;
		}
		if (po.options_array[i].nbr == onKilobyteReal) {
			oKilobyteReal = po.options_array[i].arg.bool_val;
			continue;
		}
		if (po.options_array[i].nbr == onPadding) {
			if (po.options_array[i].arg.int_val < 1 || po.options_array[i].arg.int_val > 16) {
				stringstream err;
				err << "main(): Invalid amount for the \"padding\" option.";
				err << "  Must be between 1 and 16.";
				throw runtime_error(err.str());
			}
			oPadding = po.options_array[i].arg.int_val;
			continue;
		}
		if (po.options_array[i].nbr == onReverseSort) {
			oReverseSort = po.options_array[i].arg.bool_val;
			continue;
		}
		if (po.options_array[i].nbr == onVersion) {
			oVersion = true;
			continue;
		}
		if (po.options_array[i].nbr == onSort) {
			oSort = po.options_array[i].arg.string_val;
			if (oSort != "date" && oSort != "d" && oSort != "name" && oSort != "n" && oSort != "size" && oSort != "s") {
				string err = "main(): Invalid argument for the sort option! Can be only \"date\", \"name\", or \"size\".";
				throw runtime_error(err);
			}
			if (oSort == "d")
				oSort = "date";
			else
				if (oSort == "n")
					oSort = "name";
				else
					if (oSort == "s")
						oSort = "size";
			continue;
		}
	}

	if (oHelp) {
		stringstream ss;
		if (oColorize)
			print_help_msg_color(ss);
		else
			print_help_msg_nocolor(ss);
		cout << ss.str();
		return 0;
	}

	if (oVersion) {
		stringstream ss;
		print_vers_msg(ss);
		cout << ss.str();
		return 0;
	}

	// Get program's path and current path
	path prog_path = argv[0];
	path init_path = ".";

	// Decide what we're listing
	if (po.arguments_array.size() == 0) {
		// Nothing on the command line, assume current path
		string tmp = init_path.full();
		file_info f = tmp;
		dump_directory(filelist, f);
	} else {
		// One item, if its a directory, dump its contents
		if (po.arguments_array.size() == 1) {
			file_info f = po.arguments_array[0];
			if (f.exists && f.is_dir)
				dump_directory(filelist, f);
			else
				filelist.push_back(f);
		} else {
			// Otherwise list the things that were passed to us.
			for (unsigned i = 0; i < po.arguments_array.size(); i++) {
				file_info f = po.arguments_array[i];
				filelist.push_back(f);
			}
		}
	}

	dump_list(filelist);

	return 0;


	
	}
	catch (runtime_error re) {
		if (oColorize)
			cerr << ct.set(ctcWhite, ctcRed, ctaBold);
		cerr << "d: runtime error in " << re.what();
		if (oColorize)
			cerr << ct.reset();
		cerr << endl;
		return 1;
	}
	catch (logic_error le) {
		if (oColorize)
			cerr << ct.set(ctcYellow, ctcRed, ctaBold);
		cerr << "d: logic error in " << le.what();
		if (oColorize)
			cerr << ct.reset();
		cerr << endl;
		return 1;
	}
}
