/*
 * parse_options.cpp
 *
 * Copyright (C) 2003 Ed Cogburn <ecogburn@xtn.net>
 *
 * This file is part of the program "d: The Directory Lister".
 *
 * "d: The Directory Lister" is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
 *
 * "d: The Directory Lister" is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with "d: The Directory Lister" in a file named "COPYING".
 * If not, visit their website at http://www.gnu.org for a copy, or request a copy by writing to:
 *     Free Software Foundation, Inc.
 *     59 Temple Place, Suite 330
 *     Boston, MA 02111-1307 USA
*/



/*
	A program using parse_options needs to declare something like this:

	po_declared_option opts[] = {
		{ 1,	poasCfgOnly,	poatString,		'\0',	"option-string-arg"	},
		{ 2,	poasOptOnly,	poatNone,		'a',	"option-no-arg"		},
		{ 3,	poasOptCfg,		poatBoolean,	'b',	"option-bool-arg"	},
		{ 4,	poasOptCfg,		poatInt,		'c',	"option-int-arg"	},
		{ 5,	poasOptCfg,		poatFloat,		'd',	"option-float-arg"	},
		{ 0,	poasVoid,		poatVoid,		'\0',	""					}
	};

	Pass this to parse_args() or parse_cfgs() then read the results from the vectors options_array and arguments_array.

	options_array will have the short or long options from the command line, or the "long options" from the config file.

	arguments_array will have non-options from the command line.  Not used when parsing a config file.

	The last entry in the array must start with 0, and must use the "void" enums as shown.

	poasOptCfg is for something that can be either an option in the configuration file or a command line option.

	poasCfgOnly is for something that can only show up in the configuration file.  It will be flagged as an error if it shows
	up on the command line (parsed by parse_args()).  These should specify a short name as '\0', as short options aren't supported
	for configuration files.

	poasOptOnly is for something that can only show up on the command line.  It will be flagged as an error if it shows
	up in the configuration file (parsed by parse_cfgs()).

	Note that there are also poatOptionalString, poatOptionalInt, and poatOptionalFloat that aren't used above.

	See "d: The Directory Lister" for a program using this class.
*/



#include "globals.hpp"
#include <cerrno>
#include <cstdlib>
#include <cstring>



//-----------------------------------------------------------------------------
void parse_options::read_int_arg(const char* dashes)
{
	char *buf;
	char *tail;
	long sint;
	bool maybe_an_int;
	static const char valid_chars[] = "0123456789+-xX";

	stringstream err;
	stringstream errsubstr;

	if (reading == portConfigs)
		errsubstr << "\"" << opt << "\" in \"" << cf.name.name_ext() << "\" at line " << lineno;
	else
		errsubstr << "\"" << dashes << opt;

	if (arg.size() == 0) {
		err << "parse_options():  option " << errsubstr.str() << " has a zero length argument!";
		throw runtime_error(err.str());
	}

	maybe_an_int = true;
	for (unsigned i = 0; i < arg.size(); i++) {
		bool is_valid = false;
		for (unsigned j = 0; j < sizeof(valid_chars); j++)
			if (arg[i] == valid_chars[j]) {
				is_valid = true;
				break;
			}
		if (is_valid == false) {
			maybe_an_int = false;
			break;
		}
	}

	if (maybe_an_int == false) {
		err << "parse_options():  option " << errsubstr.str() << " requires an integer argument!";
		throw runtime_error(err.str());
	}

	buf = new char [arg.size() + 1];
	strcpy(buf, arg.c_str());
	errno = 0;
	sint = strtol(buf, &tail, 0);

	if (!errno && tail != buf) {
		delete [] buf;
		ropt.arg.int_val = sint;
	} else {
		delete [] buf;
		err << "parse_options():  option " << errsubstr.str() << " has an unreadable integer argument!";
		throw runtime_error(err.str());
	}
}



//-----------------------------------------------------------------------------
void parse_options::read_float_arg(const char* dashes)
{
	char *buf;
	char *tail;
	double sfloat;
	bool maybe_a_float;
	static const char valid_chars[] = "0123456789+-.eE";

	stringstream err;
	stringstream errsubstr;

	if (reading == portConfigs)
		errsubstr << "\"" << opt << "\" in \"" << cf.name.name_ext() << "\" at line " << lineno;
	else
		errsubstr << "\"" << dashes << opt;

	if (arg.size() == 0) {
		err << "parse_options():  option " << errsubstr.str() << " has a zero length argument!";
		throw runtime_error(err.str());
	}

	maybe_a_float = true;
	for (unsigned i = 0; i < arg.size(); i++) {
		bool is_valid = false;
		for (unsigned j = 0; j < sizeof(valid_chars); j++)
			if (arg[i] == valid_chars[j]) {
				is_valid = true;
				break;
			}
		if (is_valid == false) {
			maybe_a_float = false;
			break;
		}
	}

	if (maybe_a_float == false) {
		err << "parse_options():  option " << errsubstr.str() << " requires a floating point argument!";
		throw runtime_error(err.str());
	}

	buf = new char [arg.size() + 1];
	strcpy(buf, arg.c_str());
	errno = 0;
	sfloat = strtod(buf, &tail);

	if (!errno && tail != buf) {
		delete [] buf;
		ropt.arg.float_val = sfloat;
	} else {
		delete [] buf;
		err << "parse_options():  option " << errsubstr.str() << " has an unreadable floating point argument!";
		throw runtime_error(err.str());
	}
}



//-----------------------------------------------------------------------------
void parse_options::read_string_arg(const char* dashes)
{
	stringstream err;
	stringstream errsubstr;

	if (reading == portConfigs)
		errsubstr << "\"" << opt << "\" in \"" << cf.name.name_ext() << "\" at line " << lineno;
	else
		errsubstr << "\"" << dashes << opt;

	if (arg.size() == 0) {
		err << "parse_options():  option " << errsubstr.str() << " has a zero length argument!";
		throw runtime_error(err.str());
	}

	if (arg.size() > po_string_size) {
		err << "parse_options():  option " << errsubstr.str() << " has a string argument thats too long!";
		throw runtime_error(err.str());
	}

	strcpy(ropt.arg.string_val, arg.c_str());
}



//-----------------------------------------------------------------------------
void parse_options::validate_long_option()
{
	stringstream err;

	err << "parse_options():  option \"";
	if (reading == portOptions)
		err << "--";
	err << opt << "\"";

	for (int i = 0; dopts[i].nbr != 0; i++) {
		if (opt != dopts[i].long_name)
			continue;

		if (reading == portConfigs && dopts[i].arg_src != poasOptCfg && dopts[i].arg_src != poasCfgOnly) {
			err << " can't be used in configuration files, only on the command line!";
			throw runtime_error(err.str());
		}
		if (reading == portOptions && dopts[i].arg_src != poasOptCfg && dopts[i].arg_src != poasOptOnly) {
			err << " can't be used on the command line, only in configuration files!";
			throw runtime_error(err.str());
		}

		ropt.lineno = lineno;

		switch (dopts[i].arg_type) {
			case poatBoolean:
				ropt.nbr = dopts[i].nbr;
				if (arg == "") {
					ropt.arg.bool_val = true;
					return;
				}
				if (arg != "true" && arg != "false" && arg != "t" && arg != "f" && arg != "1" && arg != "0") {
					err << " is a boolean that can only take one of [true/false/t/f/1/0] as an argument!";
					throw runtime_error(err.str());
				}
				if (arg == "true" || arg == "t" || arg == "1")
					ropt.arg.bool_val = true;
				else
					ropt.arg.bool_val = false;
				return;

			case poatNone:
				if (arg != "") {
					err << " doesn't take an argument!";
					throw runtime_error(err.str());
				}
				ropt.nbr = dopts[i].nbr;
				return;

			case poatOptionalInt:
				ropt.nbr = dopts[i].nbr;
				if (arg != "") {
					ropt.arg_is_present = true;
					read_int_arg("--");
				} else
					ropt.arg_is_present = false;
				return;

			case poatOptionalFloat:
				ropt.nbr = dopts[i].nbr;
				if (arg != "") {
					ropt.arg_is_present = true;
					read_float_arg("--");
				} else
					ropt.arg_is_present = false;
				return;

			case poatOptionalString:
				ropt.nbr = dopts[i].nbr;
				if (arg != "") {
					ropt.arg_is_present = true;
					read_string_arg("--");
				} else
					ropt.arg_is_present = false;
				return;

			case poatInt:
				if (arg == "") {
					err << " requires an integer argument!";
					throw runtime_error(err.str());
				}
				ropt.nbr = dopts[i].nbr;
				ropt.arg_is_present = true;
				read_int_arg("--");
				return;

			case poatFloat:
				if (arg == "") {
					err << " requires a floating point argument!";
					throw runtime_error(err.str());
				}
				ropt.nbr = dopts[i].nbr;
				ropt.arg_is_present = true;
				read_float_arg("--");
				return;

			case poatString:
				if (arg == "") {
					err << " requires a string argument!";
					throw runtime_error(err.str());
				}
				ropt.nbr = dopts[i].nbr;
				ropt.arg_is_present = true;
				read_string_arg("--");
				return;

			default:
				err << " has an invalid arg_type!";
				throw runtime_error(err.str());
		}
	}

	err << " is not a supported option!";
	throw runtime_error(err.str());
}



//-----------------------------------------------------------------------------
void parse_options::validate_short_option()
{
	stringstream err;
	err << "parse_options():  option \"-" << opt << "\"";

	for (int i = 0; dopts[i].nbr != 0; i++) {
		if (opt[0] != dopts[i].char_name)
			continue;

		if (dopts[i].arg_src != poasOptCfg && dopts[i].arg_src != poasOptOnly)
			continue;

		switch (dopts[i].arg_type) {
			case poatBoolean:
				ropt.nbr = dopts[i].nbr;
				if (arg.size() == 0 || (arg.size() == 1 && arg[0] == '+')) {
					ropt.arg.bool_val = true;
					return;
				}
				if (arg.size() == 1 && arg[0] == '-') {
					ropt.arg.bool_val = false;
					return;
				}
				err << " is a boolean that can only take a '+' or '-' as an argument!";
				throw runtime_error(err.str());
				return;

			case poatNone:
				if (arg == "")
					ropt.nbr = dopts[i].nbr;
				else
					err << " doesn't take an argument!";
				return;

			case poatOptionalInt:
				ropt.nbr = dopts[i].nbr;
				if (arg != "") {
					ropt.arg_is_present = true;
					read_int_arg("-");
				} else
					ropt.arg_is_present = false;
				return;

			case poatOptionalFloat:
				ropt.nbr = dopts[i].nbr;
				if (arg != "") {
					ropt.arg_is_present = true;
					read_float_arg("-");
				} else
					ropt.arg_is_present = false;
				return;

			case poatOptionalString:
				ropt.nbr = dopts[i].nbr;
				if (arg != "") {
					ropt.arg_is_present = true;
					read_string_arg("-");
				} else
					ropt.arg_is_present = false;
				return;

			case poatInt:
				if (arg == "") {
					err << " requires an integer argument!";
					throw runtime_error(err.str());
				} else {
					ropt.nbr = dopts[i].nbr;
					ropt.arg_is_present = true;
					read_int_arg("-");
				}
				return;

			case poatFloat:
				if (arg == "") {
					err << " requires a floating point argument!";
					throw runtime_error(err.str());
				} else {
					ropt.nbr = dopts[i].nbr;
					ropt.arg_is_present = true;
					read_float_arg("-");
				}
				return;

			case poatString:
				if (arg == "") {
					err << " requires a string argument!";
					throw runtime_error(err.str());
				} else {
					ropt.nbr = dopts[i].nbr;
					ropt.arg_is_present = true;
					read_string_arg("-");
				}
				return;

			default:
				err << " has an invalid arg_type!";
				throw runtime_error(err.str());
		}
	}

	err << " is not a supported option!";
	throw runtime_error(err.str());
}



//-----------------------------------------------------------------------------
void parse_options::parse_long_option(string& src)
{
	stringstream err;

	err << "parse_options():  Invalid syntax in ";
	if (reading == portConfigs)
		err << "\"" << cf.name.name_ext() << "\" at line " << lineno << '!';
	else
		err << "statement \"" << src << "\"!";

	opt = "";
	arg = "";

	unsigned idx = src.find('=', 0);
	if (idx == src.npos) {
		unsigned idx2 = src.find(' ', idx);
		if (idx2 != src.npos) {
			err << "  Spaces in the expression?";
			throw runtime_error(err.str());
		} else {
			opt = src;
			validate_long_option();
			options_array.push_back(ropt);
			return;
		}
	}

	if (idx == src.size() - 1) {
		err << "  An equals sign with nothing after it?";
		throw runtime_error(err.str());
	}

	unsigned idx2 = src.find('=', idx + 1);
	if (idx2 != src.npos) {
		err << "  More than one equals sign on the line?";
		throw runtime_error(err.str());
	}

	opt = src.substr(0, idx);
	arg = src.substr(idx + 1);

	while (opt[opt.size() - 1] == ' ' || opt[opt.size() - 1] == '\t')
		opt.erase(opt.size() - 1, 1);

	while (arg[0] == ' ' || arg[0] == '\t')
		arg.erase(0, 1);

	validate_long_option();
	options_array.push_back(ropt);
}



//-----------------------------------------------------------------------------
void parse_options::parse_short_option(string& src)
{
	stringstream err;

	err << "parse_options():  Invalid syntax in statement \"" << src << "\"!";

	opt = src[0];

	arg = "";

	if (src.size() > 1) {
		arg = src.substr(1);

		if (arg[0] == '=') {
			if (arg.size() == 1) {
				err << "  An equals sign with nothing after it?";
				throw runtime_error(err.str());
			}
			arg.erase(0, 1);
		}

		unsigned idx = arg.find(' ', 0);
		if (idx != src.npos) {
			err << "  Spaces in the expression?";
			throw runtime_error(err.str());
		}
	}

	validate_short_option();
	options_array.push_back(ropt);
}



//-----------------------------------------------------------------------------
void parse_options::parse_args(int argc, char** argv, po_declared_option* opts)
{
	string src;

	dopts = opts;

	options_array.clear();
	arguments_array.clear();

	reading = portOptions;

	if (argc < 2)
		return;

	for (int i = 1; i < argc; i++) {
		src = argv[i];

		ropt.nbr = 0;
		ropt.arg_is_present = false;
		ropt.arg.int_val = 0;

		if (src.size() == 0)
			throw runtime_error("parse_options():  a zero length option?!?");

		if (src[0] != '-') {
			arguments_array.push_back(src);
			continue;
		}

		if (src.size() == 1)
			throw runtime_error("parse_options():  a dash with no option after it!");

		if (src[1] == '-') {
			if (src.size() == 2)
				throw runtime_error("parse_options():  dashes with no option after them!");
			else {
				src.erase(0, 2);
				parse_long_option(src);
			}
		} else {
			src.erase(0, 1);
			parse_short_option(src);
		}
	}
}



//-----------------------------------------------------------------------------
void parse_options::parse_cfgs_readcfg()
{
	char buf[256];
	char* ptr = buf;
	string src;

	ifstream ifile(cf.name.full().c_str());

	while (ifile.good()) {
		ifile.getline(buf, 256, '\n');

		if (ifile.bad()) {
			stringstream ss;
			ss << "parse_options():  Can't read \"" << cf.name.name_ext() << "\" at line " << lineno << '!';
			throw runtime_error(ss.str());
		}

		lineno++;

		while (*ptr == ' ' || *ptr == '\t')
			ptr++;

		if (*ptr == '#' || *ptr == '\0')
			continue;

		src = ptr;

		for (int i = 0; src[i] != '\0'; i++)
			if (src[i] >= 'A' && src[i] <= 'Z')
				src[i] += 32;

		while (src[src.size() - 1] == ' ' || src[src.size() - 1] == '\t')
			src.erase(src.size() - 1, 1);

		parse_long_option(src);
	}
}



//-----------------------------------------------------------------------------
void parse_options::parse_cfgs(char* file, po_declared_option* opts)
{
	dopts = opts;
	cf = file;

	options_array.clear();
	arguments_array.clear();

	reading = portConfigs;

	lineno = 0;

	if (cf.exists)
		parse_cfgs_readcfg();
}



//-----------------------------------------------------------------------------
void parse_options::parse_cfgs(string file, po_declared_option* opts)
{
	dopts = opts;
	cf = file;

	options_array.clear();
	arguments_array.clear();

	reading = portConfigs;

	lineno = 0;

	if (cf.exists)
		parse_cfgs_readcfg();
}
